import {loadGLTF} from "/src/libs/loader.js";
import * as THREE from '/src/libs/three.js-r132/build/three.module.js';
import {ARButton} from '/src/libs/three.js-r132/examples/jsm/webxr/ARButton.js';

const normalizeModel = (obj, height) => {
  // scale it according to height
  const bbox = new THREE.Box3().setFromObject(obj);
  const size = bbox.getSize(new THREE.Vector3());
  obj.scale.multiplyScalar(height / size.y);

  // reposition to center
  const bbox2 = new THREE.Box3().setFromObject(obj);
  const center = bbox2.getCenter(new THREE.Vector3());
  obj.position.set(-center.x, -center.y, -center.z);
}

// recursively set opacity
const setOpacity = (obj, opacity) => {
  obj.children.forEach((child) => {
    setOpacity(child, opacity);
  });
  if (obj.material) {
    obj.material.format = THREE.RGBAFormat; // required for opacity
    obj.material.opacity = opacity;
  }
}

// make clone object not sharing materials
const deepClone = (obj) => {
  const newObj = obj.clone();
  newObj.traverse((o) => {
    if (o.isMesh) {
      o.material = o.material.clone();
    }
  });
  return newObj;
}

document.addEventListener('DOMContentLoaded', () => {
  const initialize = async() => {
    const scene = new THREE.Scene();
    const camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 0.01, 20);

    const light = new THREE.HemisphereLight( 0xffffff, 0xbbbbff, 1 );
    scene.add(light);

    const renderer = new THREE.WebGLRenderer({antialias: true, alpha: true});
    renderer.setPixelRatio(window.devicePixelRatio);
    renderer.setSize(window.innerWidth, window.innerHeight);
    renderer.xr.enabled = true;

    const arButton = ARButton.createButton(renderer, {requiredFeatures: ['hit-test'], optionalFeatures: ['dom-overlay'], domOverlay: {root: document.body}});
    document.body.appendChild(renderer.domElement);
    document.body.appendChild(arButton);


    const model = await loadGLTF('/assets/models/cat-tower.glb');
    normalizeModel(model.scene, 2.0);
    const item = new THREE.Group();
    item.add(model.scene);
    item.visible = false;
    const spawnItem = deepClone(item)
    setOpacity(item, 1.0);
    scene.add(spawnItem);
    

    let selectedItem = item;
    let prevTouchPosition = null;

    const controller = renderer.xr.getController(0);
    scene.add(controller);
    controller.addEventListener('select', () => {
      const spawnItem = deepClone(selectedItem);
      setOpacity(spawnItem, 1.0);
      scene.add(spawnItem);
    });
    
    renderer.xr.addEventListener("sessionstart", async (e) => {
      const session = renderer.xr.getSession();
      const viewerReferenceSpace = await session.requestReferenceSpace("viewer");
      const hitTestSource = await session.requestHitTestSource({space: viewerReferenceSpace});

      renderer.setAnimationLoop((timestamp, frame) => {
	if (!frame) return;

	if (selectedItem) {
	  const hitTestResults = frame.getHitTestResults(hitTestSource);
	  if (hitTestResults.length) {
	    const hit = hitTestResults[0];
      const referenceSpace = renderer.xr.getReferenceSpace();
      const viewerMatrix = new THREE.Matrix4().fromArray(frame.getViewerPose(referenceSpace).transform.inverse.matrix);
	  const newPosition = controller.position.clone();
	  newPosition.applyMatrix4(viewerMatrix); // change to viewer coordinate
	  
	  prevTouchPosition = newPosition;
	    selectedItem.visible = true;
	    selectedItem.position.setFromMatrixPosition(new THREE.Matrix4().fromArray(hit.getPose(referenceSpace).transform.matrix));
	  } else {
	    selectedItem.visible = false;
	  }
	}

	renderer.render(scene, camera);
      });
    });
  }
  initialize();
});
